import io
import logging
from typing import TYPE_CHECKING, Dict, Iterator, Optional, Union

import requests
from requests.adapters import HTTPAdapter

from odp.tabular_v2.util import Iter2Reader
from odp.util.util import size2human

if TYPE_CHECKING:
    from odp.tabular_v2.client.table import Table


class Client:
    def __init__(self, base_url: str):
        self._base_url = base_url

    class Response:
        # Abstraction for response object, shared between http client and test client
        def __init__(self, res: Union[requests.Response, Iterator[bytes], Dict, bytes]):
            if isinstance(res, requests.Response):
                if res.status_code == 204:
                    raise FileNotFoundError(res.text)
                res.raise_for_status()
            # logging.info("response: %s", res)
            self.res = res

        def reader(self):
            if isinstance(self.res, bytes):
                return io.BytesIO(self.res)
            if isinstance(self.res, Iterator):
                return Iter2Reader(self.res)
            return self.res.raw

        def iter(self) -> Iterator[bytes]:
            if isinstance(self.res, bytes):
                return iter([self.res])
            if isinstance(self.res, Iterator):
                return self.res
            return self.res.iter_content()

        def all(self) -> bytes:
            if isinstance(self.res, bytes):
                return self.res
            if isinstance(self.res, Iterator):
                return b"".join(self.res)
            return self.res.content

        def json(self) -> dict:
            if self.res is None:
                return None
            if isinstance(self.res, dict):
                return self.res
            return self.res.json()

    def _request(
        self,
        path: str,
        data: Union[Dict, bytes, None] = None,
        params: Optional[Dict] = None,
        headers: Optional[Dict] = None,
    ) -> Response:
        with requests.Session() as s:
            adapter = HTTPAdapter(pool_connections=4, pool_maxsize=4, max_retries=8)
            s.mount("http://", adapter)
            s.mount("https://", adapter)
            logging.info("sending %s %s (%s)", path, params, size2human(len(data)) if data else "-")
            if isinstance(data, dict):
                res = s.post(self._base_url + path, headers=headers, params=params, json=data, stream=True)
            elif isinstance(data, bytes):
                res = s.post(self._base_url + path, headers=headers, params=params, data=data, stream=True)
            elif isinstance(data, Iterator):
                res = s.post(self._base_url + path, headers=headers, params=params, data=data, stream=True)
            elif data is None:
                res = s.post(self._base_url + path, headers=headers, params=params, stream=True)
            else:
                raise ValueError(f"unexpected type {type(data)}")
        logging.info("got response %s", res.status_code)
        return self.Response(res)

    def table(self, table_id: str) -> "Table":
        """
        create a table handler for the given table_id
        @param table_id:
        @return:
        """
        from odp.tabular_v2.client import Table

        return Table(self, table_id)



import json
from contextlib import contextmanager
from typing import Any, Dict, Iterable, Literal, Optional, Union

import requests
import validators
from requests.adapters import HTTPAdapter

from ..util.cheapdantic import BaseModel
from .auth import TokenProvider
from .exc import OdpForbiddenError, OdpUnauthorizedError

ParamT = Optional[Dict[str, Any]]
HeaderT = Optional[Dict[str, Any]]
ContentT = Union[bytes, str, dict, list, BaseModel, None]


class OdpHttpClient(BaseModel):
    base_url: str = "https://api.hubocean.earth"
    token_provider: TokenProvider
    custom_user_agent: Optional[str] = None

    def __init__(self, base_url: str, token_provider: TokenProvider, custom_user_agent: Optional[str] = None, **kwargs):
        if not validators.url(base_url, simple_host=True):
            raise ValueError(f"Invalid base URL: {base_url}")
        base_url = base_url.rstrip("/")
        super().__init__(
            base_url=base_url, token_provider=token_provider, custom_user_agent=custom_user_agent, **kwargs
        )

    def get(
        self,
        url: str,
        params: ParamT = None,
        headers: HeaderT = None,
        timeout: Optional[float] = None,
        stream: bool = False,
    ) -> requests.Response:
        """Make a GET request

        Args:
            url: URL to request. If it starts with a "/", it will be appended to the base URL.
            params: Query parameters
            headers: HTTP headers
            timeout: Timeout in seconds
            stream: If True, the response will be streamed.

        Returns:
            The response object
        """
        return self._request("GET", url, params=params, headers=headers, timeout=timeout, stream=stream)

    def post(
        self,
        url: str,
        params: ParamT = None,
        headers: HeaderT = None,
        timeout: Optional[float] = None,
        content: ContentT = None,
        stream: bool = False,
    ) -> requests.Response:
        """Make a POST request

        Args:
            url: URL to request. If it starts with a "/", it will be appended to the base URL.
            params: Query parameters
            headers: HTTP headers
            timeout: Timeout in seconds
            content: Request body content.
            stream: If True, the response will be streamed.

        Returns:
            The response object
        """
        return self._request(
            "POST",
            url,
            params=params,
            headers=headers,
            timeout=timeout,
            content=content,
            stream=stream,
        )

    def patch(
        self,
        url: str,
        params: ParamT = None,
        headers: HeaderT = None,
        timeout: Optional[float] = None,
        content: ContentT = None,
        stream: bool = False,
    ) -> requests.Response:
        """Make a PATCH request

        Args:
            url: URL to request. If it starts with a "/", it will be appended to the base URL.
            params: Query parameters
            headers: HTTP headers
            timeout: Timeout in seconds
            content: Request body content.
            stream: If True, the response will be streamed.

        Returns:
            The response object
        """
        return self._request(
            "PATCH",
            url,
            params=params,
            headers=headers,
            timeout=timeout,
            content=content,
            stream=stream,
        )

    def put(
        self,
        url: str,
        params: ParamT = None,
        headers: HeaderT = None,
        timeout: Optional[float] = None,
        content: ContentT = None,
        stream: bool = False,
    ) -> requests.Response:
        """Make a PUT request

        Args:
            url: URL to request. If it starts with a "/", it will be appended to the base URL.
            params: Query parameters
            headers: HTTP headers
            timeout: Timeout in seconds
            content: Request body content.
            stream: If True, the response will be streamed.

        Returns:
            The response object
        """
        return self._request(
            "PUT",
            url,
            params=params,
            headers=headers,
            timeout=timeout,
            content=content,
            stream=stream,
        )

    def delete(
        self,
        url: str,
        params: ParamT = None,
        headers: HeaderT = None,
        timeout: Optional[float] = None,
        stream: bool = False,
    ) -> requests.Response:
        """Make a DELETE request

        Args:
            url: URL to request. If it starts with a "/", it will be appended to the base URL.
            params: Query parameters
            headers: HTTP headers
            timeout: Timeout in seconds
            stream: If True, the response will be streamed.

        Returns:
            The response object
        """
        return self._request(
            "DELETE",
            url,
            params=params,
            headers=headers,
            timeout=timeout,
            stream=stream,
        )

    @contextmanager
    def _session(self) -> Iterable[requests.Session]:
        """Context manager for a requests session

        Will add authentication to the session if a token provider is set.

        Yields:
            A requests session
        """
        s = requests.Session()

        # NOTE(oha): the max_retries here applies only to DNS looks and connections, not to the request itself
        # so it's safe to retry even if the request has a side effect
        # this might fix ODP-1549 where we see random connection errors, where the theory goes:
        # - the connection has been idle for a while (we can't set a TTL)
        # - k8s load balance closes after a while (I believe without sending FIN)
        # - the proxy unaware tries to reuse the connection and fails
        # this should retrie the process with another connection or a new one
        s.mount("https://", HTTPAdapter(max_retries=3))
        s.mount("http://", HTTPAdapter(max_retries=3))

        if self.token_provider:
            s.auth = self.token_provider

        yield s

    def _request(
        self,
        method: Literal["GET", "POST", "PUT", "PATCH", "DELETE"],
        url: str,
        params: ParamT = None,
        headers: HeaderT = None,
        timeout: Optional[float] = None,
        content: ContentT = None,
        stream: bool = False,
    ) -> requests.Response:
        """Make an HTTP request

        Args:
            method: HTTP method (e.g. "GET", "POST", etc.)
            url: URL to request. If it starts with a "/", it will be appended to the base URL.
            params: Query parameters
            headers: HTTP headers
            timeout: Timeout in seconds
            content: Request body content.
                If it is a dict or list, it will be serialized as JSON.
                If it is a pydantic BaseModel, it will be serialized as JSON.
            stream: If True, the response will be streamed.

        Returns:
            The response object

        Raises:
            OdpUnauthorizedError: Unauthorized request
            OdpForbiddenError: Forbidden request
        """

        if url.startswith("/"):
            base_url = self.base_url
        elif url.startswith(self.base_url):
            url = url[len(self.base_url) :]
            base_url = self.base_url
        else:
            base_url = ""

        headers = headers or {}
        if self.custom_user_agent:
            headers["User-Agent"] = self.custom_user_agent
        else:
            headers["User-Agent"] = self.token_provider.user_agent

        if isinstance(content, (dict, list)):
            body = json.dumps(content)
            headers["Content-Type"] = "application/json"
        elif isinstance(content, BaseModel):
            body = content.json().encode("utf-8")
            headers["Content-Type"] = "application/json"
        elif isinstance(content, (bytes, str)):
            body = content
            headers.setdefault("Content-Type", "application/octet-stream")
        else:
            body = None

        request_url = f"{base_url}{url}"

        req = requests.Request(method, request_url, params=params, headers=headers, data=body)

        with self._session() as s:
            prepped = s.prepare_request(req)
            res = s.send(prepped, timeout=timeout, stream=stream)

            if res.status_code == 401:
                raise OdpUnauthorizedError("Unauthorized request")
            elif res.status_code == 403:
                raise OdpForbiddenError("Forbidden request")

            return res



from typing import Optional

from odp.util.cheapdantic import BaseModel


class ContactInfo(BaseModel):
    """Contact information for a user"""

    contact: str
    """Contact in the form `Firstname Lastname <email>`"""

    organisation: Optional[str] = None
    """Organisation name"""



import json
from typing import Optional, Union

import geojson
from shapely import wkb, wkt
from shapely.geometry import shape


def convert_geometry(
    data: Union[str, dict, list, bytes], result_geometry: str, rounding_precision: Optional[int] = None
):
    if result_geometry == "wkb":
        return _convert_geometry_to_wkb(data)
    elif result_geometry == "wkt":
        return _convert_geometry_to_wkt(data, rounding_precision)
    elif result_geometry == "geojson":
        if _is_geojson(data):
            return data
        return _convert_geometry_to_geojson(data)


def _convert_geometry_to_wkb(data: Union[str, dict, list]):
    if _is_geojson(data):
        return _convert_geojson_to_wkb(data)
    if isinstance(data, str):
        try:
            return _convert_wkt_to_wkb(data)
        except Exception:
            pass
    elif isinstance(data, dict):
        for key in data:
            value = data[key]
            data[key] = _convert_geometry_to_wkb(value)
    elif isinstance(data, list):
        for i, row in enumerate(data):
            try:
                data[i] = _convert_geometry_to_wkb(row)
            except Exception:
                continue
    return data


def _convert_geometry_to_wkt(data: Union[str, dict, list, bytes], rounding_precision: Optional[int] = None):
    if _is_geojson(data):
        return _convert_geojson_to_wkt(data)
    if isinstance(data, (str, bytes)):
        try:
            return _convert_wkb_to_wkt(data, rounding_precision)
        except Exception:
            pass
    elif isinstance(data, dict):
        for key in data:
            value = data[key]
            data[key] = _convert_geometry_to_wkt(value, rounding_precision)
    elif isinstance(data, list):
        for i, row in enumerate(data):
            try:
                data[i] = _convert_geometry_to_wkt(row, rounding_precision)
            except Exception:
                continue
    return data


def _convert_geometry_to_geojson(data: Union[str, dict, list, bytes]):
    if isinstance(data, str):
        try:
            if _is_wkt(data):
                return _convert_wkt_to_geojson(data)
            else:
                return _convert_wkb_to_geojson(data)
        except Exception:
            pass
    elif isinstance(data, bytes):
        try:
            return _convert_wkb_to_geojson(data)
        except Exception:
            pass
    elif isinstance(data, dict):
        for key in data:
            value = data[key]
            data[key] = _convert_geometry_to_geojson(value)
    elif isinstance(data, list):
        for i, row in enumerate(data):
            try:
                data[i] = _convert_geometry_to_geojson(row)
            except Exception:
                continue
    return data


def _convert_geojson_to_wkb(geojson_data: Union[dict, str]) -> bytes:
    if isinstance(geojson_data, dict):
        geojson_data = json.dumps(geojson_data)
    geo = geojson.loads(geojson_data)
    return shape(geo).wkb


def _convert_geojson_to_wkt(geojson_data: Union[dict, str]) -> str:
    if isinstance(geojson_data, dict):
        geojson_data = json.dumps(geojson_data)
    geo = geojson.loads(geojson_data)
    return shape(geo).wkt


def _convert_wkb_to_geojson(wkb_data: Union[bytes, str]) -> dict:
    geo = wkb.loads(wkb_data)
    return geojson.Feature(geometry=geo, properties={}).geometry


def _convert_wkb_to_wkt(wkb_data: Union[bytes, str], rounding_precision: Optional[int] = None) -> str:
    if rounding_precision:
        return wkt.dumps(wkb.loads(wkb_data), rounding_precision=rounding_precision)
    return wkt.dumps(wkb.loads(wkb_data))


def _convert_wkt_to_geojson(wkt_data: str) -> dict:
    geo = wkt.loads(wkt_data)
    return geojson.Feature(geometry=geo, properties={}).geometry


def _convert_wkt_to_wkb(wkt_data: str) -> bytes:
    return wkb.dumps(wkt.loads(wkt_data))


def _is_geojson(data) -> bool:
    if isinstance(data, dict):
        return len(data.keys()) == 2 and "type" in data and "coordinates" in data
    elif isinstance(data, str):
        try:
            return _is_geojson(json.loads(data))
        except Exception:
            return False
    return False


def _is_wkt(data: str) -> bool:
    # Cheap way of checking if the value is a WKT string
    #   Simply see if the first character is the first letter of a WKT-Object:
    #     P: Point, Polygon
    #     L: LineString
    #     M: MultiPoint, MultiPolygon, MultiLineString
    #     G: GeometryCollection
    return data[0].upper() in {"P", "L", "M", "G"}



from .catalog import *  # noqa: F401, F403
from .metadata import MetadataDto
from .resource import ResourceDto, ResourceSpec
from .resource_registry import DEFAULT_RESOURCE_TYPE_REGISTRY
from .resource_status import ResourceStatus



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional
from uuid import UUID

from odp.util.cheapdantic import BaseModel, Field


class Visibility(str, Enum):
    PUBLIC = "public"
    PRIVATE = "private"


class PublishStatus(str, Enum):
    PUBLISHED = "published"
    DRAFT = "draft"
    IN_REVIEW = "in_review"
    ARCHIVED = "archived"


class LicenseEntryType(str, Enum):
    CUSTOM = "custom"
    ENUM = "enum"


class ContributorEntryType(str, Enum):
    CURATED = "curated"
    CUSTOM = "custom"


class ContributorKind(str, Enum):
    PERSON = "person"
    ORG = "org"


class DatasetPropertyType(str, Enum):
    STATIC = "static"
    COMPUTED = "computed"


class ContributorRole(str, Enum):
    PUBLISHER = "publisher"
    AGGREGATOR = "aggregator"
    MAINTAINER = "maintainer"
    CREATOR = "creator"
    RIGHTS_HOLDER = "rights_holder"
    CONTRIBUTOR = "contributor"


class AuditFields(BaseModel):
    created_by: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_by: Optional[str] = None
    updated_at: Optional[datetime] = None


class Constraint(BaseModel):
    text: str


class Citation(BaseModel):
    text: str
    link: str


class Aggregation(BaseModel):
    aggregator: str
    property: str


class DatasetProperty(BaseModel):
    name: str
    description: str
    type: DatasetPropertyType
    value: str
    aggregations: List[Aggregation] = Field(default_factory=list)
    unit: str
    annotations: Dict[str, str] = Field(default_factory=dict)


class LicenseEntity(BaseModel):
    id: Optional[UUID] = None
    name: Optional[str] = None
    enum_val: Optional[str] = None
    entry_type: LicenseEntryType = None  # custom | enum
    text: Optional[str] = None
    url: Optional[str] = None
    audit_fields: Optional[AuditFields] = None


class Contributor(BaseModel):
    id: UUID
    name: str
    description: str
    website: str
    email: str
    parent_id: Optional[UUID] = None
    kind: ContributorKind
    entry_type: ContributorEntryType
    audit_fields: Optional[AuditFields] = None


class DataCollection(BaseModel):
    id: Optional[UUID] = None
    name: Optional[str] = None
    description: Optional[str] = None
    owner: Optional[str] = None
    visibility: Visibility
    deleted: bool = False
    datasets: List["Dataset"] = Field(default_factory=list)
    audit_fields: Optional[AuditFields] = None


class Dataset(BaseModel):
    id: Optional[UUID] = None
    name: Optional[str] = None
    description: Optional[str] = None
    provider: Optional[Contributor] = None
    constraints: List[Constraint] = Field(default_factory=list)
    citations: List[Citation] = Field(default_factory=list)
    contributors: List[Contributor] = Field(default_factory=list)
    license: Optional[LicenseEntity] = None
    lineage: Optional[Dict[str, List["Dataset"]]] = None
    tags: List[str] = Field(default_factory=list)
    visibility: Optional[str] = None
    deleted: bool = False
    read_only: bool = False
    publish_status: PublishStatus
    published_at: Optional[datetime] = None
    properties: List[DatasetProperty] = Field(default_factory=list)
    owner: Optional[str] = None
    collection: Optional[DataCollection] = None
    audit_fields: AuditFields


class CustomContributor(BaseModel):
    name: str
    description: str
    website: str
    email: str
    kind: str


class DatasetContributorRequest(BaseModel):
    contributor_id: str


class DatasetLicenseDto(BaseModel):
    name: str
    text: str
    url: str


class CreateDatasetRequest(BaseModel):
    name: str
    description: str
    provider_id: Optional[str] = None
    custom_provider: Optional[CustomContributor] = None
    constraints: List[Constraint] = Field(default_factory=list)
    citations: List[Citation] = Field(default_factory=list)
    contributors: Dict[str, List[DatasetContributorRequest]] = Field(default_factory=dict)
    custom_license: Optional[DatasetLicenseDto] = None
    license_enum: Optional[str] = None
    lineage: Dict[str, List[str]] = Field(default_factory=dict)
    tags: List[str] = Field(default_factory=list)


class UpdateDatasetRequest(CreateDatasetRequest):
    name: str
    description: str
    provider_id: Optional[str] = None
    custom_provider: Optional[CustomContributor] = None
    constraints: List[Constraint] = Field(default_factory=list)
    citations: List[Citation] = Field(default_factory=list)
    contributors: Dict[str, List[DatasetContributorRequest]] = Field(default_factory=dict)
    custom_license: Optional[DatasetLicenseDto] = None
    license_enum: Optional[str] = None
    lineage: Dict[str, List[str]] = Field(default_factory=dict)
    tags: List[str] = Field(default_factory=list)


class CreateDataCollectionRequest(BaseModel):
    name: str
    description: str


class UpdateDataCollectionRequest(BaseModel):
    name: str
    description: str


class DatasetPermissionStates(BaseModel):
    visibility: str
    read_only: bool
    publish_status: str


class DataCollectionPermissionStates(BaseModel):
    visibility: str



import datetime
from typing import Dict, List, Union


def vars_to_json(vars: Union[Dict, List, None]) -> Union[Dict, List[Dict], None]:
    """
    Convert a dictionary of variables to a JSON-serializable dictionary which preserve datetime type
    """
    if not vars:
        return None

    def _var_to_json(value):
        if isinstance(value, str):
            return {"type": "string", "value": value}
        if isinstance(value, int):
            return {"type": "integer", "value": value}
        if isinstance(value, float):
            return {"type": "float", "value": value}
        if isinstance(value, bool):
            return {"type": "boolean", "value": value}
        if isinstance(value, datetime.datetime):
            return {"type": "datetime", "value": value.isoformat()}
        if isinstance(value, datetime.date):
            return {"type": "date", "value": value.isoformat()}
        if isinstance(value, datetime.time):
            return {"type": "time", "value": value.isoformat()}
        else:
            raise ValueError(f"unsupported type {type(value)}")

    if isinstance(vars, dict):
        return {k: _var_to_json(v) for k, v in vars.items()}
    elif isinstance(vars, list):
        return [_var_to_json(v) for v in vars]
    else:
        raise ValueError(f"unsupported type {type(vars)}")


def json_to_vars(j: Union[Dict, List[Dict], None]) -> Union[Dict, List, None]:
    """
    Convert back a JSON dict to variables
    """
    if not j:
        return None

    def _json_to_var(j):
        if j["type"] == "string":
            return j["value"]
        if j["type"] == "integer":
            return int(j["value"])
        if j["type"] == "float":
            return j["value"]
        if j["type"] == "boolean":
            return j["value"]
        if j["type"] == "datetime":
            return datetime.datetime.fromisoformat(j["value"])
        if j["type"] == "date":
            return datetime.date.fromisoformat(j["value"])
        if j["type"] == "time":
            return datetime.time.fromisoformat(j["value"])
        else:
            raise ValueError(f"unsupported type {j['type']}")

    if isinstance(j, list):
        return [_json_to_var(v) for v in j]
    elif isinstance(j, dict):
        return {k: _json_to_var(v) for k, v in j.items()}
    else:
        raise ValueError(f"unsupported type {type(j)}")



import datetime
from decimal import Decimal
from typing import Dict, List, Union

import pandas as pd
import pyarrow as pa
from pyarrow import Field


def validate_data_against_schema(data: Union[List[Dict], pa.RecordBatch, pd.DataFrame], schema: pa.Schema):
    """
    Validate data against schema
    Checks for required fields and data types
    """
    assert schema is not None
    assert data is not None

    if isinstance(data, list):
        for i, row in enumerate(data):
            if not isinstance(row, dict):
                raise ValueError(f"expected dict but got {type(row)}")
            _check_required_fields(list(row.keys()), schema)
            for k, v in row.items():
                try:
                    field = schema.field(k)
                    _validate_data_cell(i, field, v, k)
                except KeyError:
                    pass

    elif isinstance(data, pa.RecordBatch):
        _check_required_fields(data.schema.names, schema)
        for field in data.schema:
            column_data = data.column(field.name)
            for i in range(data.num_rows):
                py_val = column_data[i].as_py()
                _validate_data_cell(i, field, py_val, field.name)

    elif isinstance(data, pd.DataFrame):
        _check_required_fields(data.columns, schema)
        for column_name in data.columns:
            field = schema.field(column_name)
            data[column_name].apply(lambda cell: _validate_data_cell(0, field, cell, column_name))

    else:
        raise ValueError(f"unexpected type {type(data)}")


def _check_required_fields(input_field_names: List[str], schema: pa.Schema):
    for field_name in schema.names:
        field = schema.field(field_name)
        if field_name not in input_field_names and field.nullable is False:
            raise ValueError(f"missing required column: {field_name}")


def _validate_data_cell(row_index: int, field: Field, v: any, k: str):
    if v is None:
        if field.nullable is False:
            raise ValueError(f"missing required column value at row index {row_index}: {field.name}")
        else:
            return

    if pa.types.is_string(field.type):
        if not isinstance(v, str):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_integer(field.type):
        if not isinstance(v, int):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_floating(field.type):
        if not isinstance(v, (int, float)):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_boolean(field.type):
        if not isinstance(v, bool):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_date(field.type):
        if not isinstance(v, datetime.date):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_timestamp(field.type):
        if not isinstance(v, datetime.datetime):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_time(field.type):
        if not isinstance(v, datetime.time):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_decimal(field.type):
        if not isinstance(v, int) and not isinstance(v, Decimal):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")
    if pa.types.is_binary(field.type):
        if not isinstance(v, bytes):
            raise ValueError(f"unexpected type {type(v)} for {k} at row index {row_index}")



from typing import TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    from odp.tabular_v2.client import Table


class Raw:
    def __init__(self, table: "Table"):
        self.table = table

    def list(self) -> list:
        res = self.table._client._request(
            "/api/table/v2/raw/list",
            params={"table_id": self.table._id},
            data={},
        )
        body = res.json()
        return body["files"]

    def upload(self, name: str, data: bytes) -> str:
        res = self.table._client._request(
            "/api/table/v2/raw/upload",
            params={"table_id": self.table._id, "name": name},
            data=data,
        )
        body = res.json()
        return body["raw_id"]

    def download(self, raw_id: str) -> Iterator[bytes]:
        res = self.table._client._request(
            "/api/table/v2/raw/download",
            params={"table_id": self.table._id, "raw_id": raw_id},
        )
        return res.iter()

    def ingest(self, raw_id: str) -> dict:
        res = self.table._client._request(
            "/api/table/v2/raw/ingest",
            params={"table_id": self.table._id, "raw_id": raw_id},
            data={},
        )
        body = res.json()
        return body



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, Generic, List, Optional, Set, Type, TypeVar

if TYPE_CHECKING:
    from odp.dto.resource import ResourceDto

T = TypeVar("T")


class TreeNode(Generic[T], ABC):
    def __init__(self, key: str):
        self.key = key
        self.data: Optional[T] = None
        self.children: Dict[str, TreeNode] = {}

    @abstractmethod
    def add_data(self, data: T) -> None:
        """Add a data to the node"""

    @abstractmethod
    def get_data(self, *args, **kwargs) -> T:
        """Get the resource"""

    def get_child(self, child: str) -> "TreeNode":
        return self.children.get(child)

    def add_child(self, child: str) -> "TreeNode":
        return self.children.setdefault(child, self.__class__(key=child))


class TreeStructure(Generic[T]):
    def __init__(self):
        self.root: Optional[ResourceNode] = None

    @classmethod
    @abstractmethod
    def add(cls, *args, **kwargs) -> None:
        """Add a new element to the tree"""

    @abstractmethod
    def get(self, *args, **kwargs) -> T:
        """Get the resource"""

    def add_leaf(self, levels) -> TreeNode:
        assert self.root is not None
        node: TreeNode = self.root
        for level in levels:
            node = node.add_child(level)
        return node

    def get_leaf(self, levels: List[str]) -> Optional[TreeNode]:
        node = self.root
        for level in levels:
            node = node.get_child(level)
            if node is None:
                return None
        return node


class ResourceNode(TreeNode[Dict[str, Type["ResourceDto"]]]):  # noqa: F821
    def __init__(self, key: str):
        super().__init__(key=key)
        self.data = {}

    def add_data(self, resource_type: Type["ResourceDto"]):  # noqa: F821
        if self.data.get(resource_type.get_version()):
            raise ValueError(f"Resource version '{resource_type.get_version()}' already exists")
        self.data[resource_type.get_version()] = resource_type

    def get_data(self, version: str) -> Optional[Type["ResourceDto"]]:  # noqa: F821
        return self.data.get(version)

    @property
    def versions(self) -> Set[str]:
        return set(self.data.keys())


class ResourceTypeRegistry(TreeStructure[Dict[str, Type["ResourceDto"]]]):  # noqa: F821
    def __init__(self) -> None:
        super().__init__()
        self.root = ResourceNode(key="")

    def add(self, resource_type: Type["ResourceDto"]) -> None:  # noqa: F821
        node = self.add_leaf(resource_type.get_kind().split("/"))
        node.add_data(resource_type)

    def get(
        self, kind: str, version: str, default: Optional[Type["ResourceDto"]] = None  # noqa: F821
    ) -> Optional[Type["ResourceDto"]]:  # noqa: F821
        node: Optional[ResourceNode] = self.get_leaf(kind.split("/"))
        if node is None:
            return default
        resource = node.get_data(version)
        if resource is None:
            raise ValueError(f"Version '{version}' not found for kind '{kind}'. Use a valid version: {node.versions}")
        return resource


DEFAULT_RESOURCE_TYPE_REGISTRY = ResourceTypeRegistry()



# dto_to_stac.py
import logging
import os.path
from datetime import datetime, timezone
from typing import Any, Dict, List

import pystac
import shapely

from odp.client import OdpClient
from odp.dto import DataCollectionDto, DatasetDto, ObservableDto, ResourceDto
from odp.dto.registry.observables_class_definitions import (
    static_coverage_class,
    static_geometric_coverage_class,
    static_single_value_class,
)
from odp.stac import CATALOG_FRONTEND_URL, STAC_API_URL


def get_root_catalog() -> pystac.Catalog:
    root_catalog = pystac.Catalog(id="stac-root", description="Hub Ocean Public STAC API", href=STAC_API_URL)

    # Collections link
    root_catalog.add_link(
        pystac.Link(
            rel="data",
            target=os.path.join(STAC_API_URL, "collections"),
            media_type=pystac.MediaType.JSON,
        )
    )
    # Search link
    root_catalog.add_link(
        pystac.Link(
            rel="search",
            target=os.path.join(STAC_API_URL, "search"),
            media_type=pystac.MediaType.GEOJSON,
        )
    )

    # Link to webview for catalog
    root_catalog.add_link(
        pystac.Link(
            rel=pystac.RelType.ALTERNATE,
            target=CATALOG_FRONTEND_URL,
            media_type=pystac.MediaType.HTML,
            title="Hub Ocean Web Catalog",
        )
    )
    return root_catalog


def get_observables(dto: ResourceDto, odp_client: OdpClient) -> List[ObservableDto]:
    """
    Get observables associated with a dataset.
    """
    try:
        observables = odp_client.catalog.list(
            {"#AND": [{"#EQUALS": ["kind", ObservableDto.get_kind()]}, {"#EQUALS": ["$spec.ref", dto.qualified_name]}]}
        )
    except Exception as e:
        logging.info(f"Error fetching observables for dataset {dto.uuid}: {e}")
        observables = []

    return list(observables)


class ObservablesInfo:
    start_time: datetime = None
    end_time: datetime = None
    geometry: Dict[str, Any] = None
    bbox: List[float] = None
    time: datetime = None

    def add_observable(self, observable: ObservableDto):
        if observable.spec.observable_class == static_coverage_class.qualified_name:
            try:
                self.start_time = datetime.fromtimestamp(observable.spec.details["value"][0])
                self.end_time = datetime.fromtimestamp(observable.spec.details["value"][1])
            except ValueError as e:
                logging.info(f"Error parsing observable {observable.uuid}: {e}")
                raise
        elif observable.spec.observable_class == static_single_value_class.qualified_name:
            try:
                value = observable.spec.details["value"]
                if isinstance(value, (int, float)):
                    self.time = datetime.fromtimestamp(value)
                else:
                    self.time = datetime.fromisoformat(value)
            except (ValueError, TypeError) as e:
                logging.info(f"Error parsing observable {observable.uuid}: {e}")
                raise ValueError(f"Error parsing observable {observable.uuid}: {e}")
        elif observable.spec.observable_class == static_geometric_coverage_class.qualified_name:
            try:
                self.geometry = observable.spec.details["value"]
                self.bbox = shapely.geometry.shape(self.geometry).bounds
            except shapely.errors.ShapelyError as e:
                logging.info(f"Error parsing observable {observable.uuid}: {e}")
                raise
        else:
            raise ValueError(f"Observable class {observable.spec.observable_class} not supported")

    def get_default_time(self) -> datetime:
        """Returns default time to current time if no time is available."""
        if self.start_time and self.end_time:
            return self.time
        elif self.time is None:
            return datetime.now(tz=timezone.utc)
        return self.time

    def get_temporal_extent(self) -> pystac.TemporalExtent:
        """Returns a pystac.TemporalExtent object based on the available time information."""
        if self.start_time and self.end_time:
            return pystac.TemporalExtent(intervals=[[self.start_time, self.end_time]])
        elif self.time:
            return pystac.TemporalExtent(intervals=[[self.time, None]])
        else:
            return pystac.TemporalExtent(intervals=[None, None])

    def get_spatial_extent(self) -> pystac.SpatialExtent:
        """Returns a pystac.SpatialExtent object based on the available spatial information.

        Defaults to the global extent if no spatial information is available.
        """
        if self.bbox:
            return pystac.SpatialExtent(bboxes=[self.bbox])
        return pystac.SpatialExtent(bboxes=[[-180, -90, 180, 90]])

    @classmethod
    def from_observables(cls, observables: List[ObservableDto]) -> "ObservablesInfo":
        info = cls()
        for observable in observables:
            try:
                info.add_observable(observable)
            except ValueError:
                continue
        return info


def convert_dataset(dto: DatasetDto) -> pystac.Item:
    """
    Converts a Dataset DTO to a STAC Item.
    """
    observables_info = ObservablesInfo.from_observables(dto.spec.observables)

    if dto.spec.data_collection is None:
        # Raise an error if the dataset does not belong to a collection. The
        # reason for this is that this stac api implementation assumes that all
        # items follow the format /collections/{collection_id}/items/{feature_id}
        raise AssertionError(f"Dataset does not belong to a collection: {dto}")

    dataset_identifier = dto.metadata.name
    collection_identifier = dto.spec.data_collection.split("/")[-1]

    dataset_api_url = os.path.join(STAC_API_URL, "collections", collection_identifier, "items", dataset_identifier)

    # Create a STAC Item
    item = pystac.Item(
        stac_extensions=[],
        href=dataset_api_url,
        id=str(dto.uuid),
        geometry=observables_info.geometry,
        bbox=observables_info.bbox,
        properties={},
        datetime=observables_info.get_default_time(),
        start_datetime=observables_info.start_time,
        end_datetime=observables_info.end_time,
        assets={},
        collection=dto.spec.data_collection,
    )

    # Add links

    # Update self link to make sure it has the correct media type
    item.remove_links("self")
    item.add_link(pystac.Link(rel="self", target=dataset_api_url, media_type=pystac.MediaType.GEOJSON))

    # Set root object (this is to get the correct root link)
    item.set_root(get_root_catalog())

    # Add link to parent collection
    item.add_link(
        pystac.Link(
            rel=pystac.RelType.COLLECTION,
            target=os.path.join(STAC_API_URL, "collections", collection_identifier),
            title=collection_identifier,
            media_type=pystac.MediaType.JSON,
        )
    )

    # Link to webview for catalog
    item.add_link(
        pystac.Link(
            rel=pystac.RelType.ALTERNATE,
            target=os.path.join(CATALOG_FRONTEND_URL, "dataset", dataset_identifier),
            media_type=pystac.MediaType.HTML,
            title="Hub Ocean Web Catalog",
        )
    )

    # Validate STAC Item
    try:
        item.validate()
    except pystac.STACValidationError as e:
        logging.warning("Validation failed for DatasetDto %s converted to a STAC Item %s: %s", dto.uuid, item, e)
        raise

    return item


def convert_collection(dto: DataCollectionDto) -> pystac.Collection:
    """
    Converts a DataCollection DTO to a STAC Collection.
    """
    observables_info = ObservablesInfo.from_observables(dto.spec.observables)

    extent = pystac.Extent(
        spatial=observables_info.get_spatial_extent(),
        temporal=observables_info.get_temporal_extent(),
    )

    # Determine the license name (default to "proprietary")
    license_name = "proprietary"
    license_href = None

    if dto.spec.distribution and dto.spec.distribution.license:
        license_obj = dto.spec.distribution.license  # This is an instance of License

        def normalize_license(li_name):
            li_name = li_name.replace(" ", "_")
            li_name = li_name.replace("(", "")
            li_name = li_name.replace(")", "")
            return li_name

        license_name = normalize_license(license_obj.name or license_name)  # License name
        license_href = license_obj.href  # URL to license text

    collection_api_url = os.path.join(STAC_API_URL, "collections", dto.metadata.name)

    # Create pystac.Collection
    collection = pystac.Collection(
        id=str(dto.uuid),
        description=dto.metadata.description or "",
        href=collection_api_url,
        title=dto.metadata.name,
        license=license_name,
        extent=extent,
    )

    # Add links for collection
    # Set root object (this is to get the correct root link)
    collection.set_root(get_root_catalog())

    # Parent link
    collection.add_link(
        pystac.Link(
            rel=pystac.RelType.PARENT,
            target=os.path.join(STAC_API_URL, "collections"),
            media_type=pystac.MediaType.JSON,
        )
    )

    # Items link
    collection.add_link(
        pystac.Link(
            rel=pystac.RelType.ITEMS,
            target=os.path.join(collection_api_url, "items"),
            media_type=pystac.MediaType.GEOJSON,
        )
    )

    # Link to webview for catalog
    collection.add_link(
        pystac.Link(
            rel=pystac.RelType.ALTERNATE,
            target=os.path.join(CATALOG_FRONTEND_URL, "collection", str(dto.metadata.name)),
            media_type=pystac.MediaType.HTML,
            title="Hub Ocean Web Catalog",
        )
    )

    # Add license link if href is available
    if license_href:
        collection.add_link(
            pystac.Link(
                rel="license",
                target=license_href,
                title=f"License: {license_name}",
            )
        )

    # Validate STAC collection
    try:
        collection.validate()
    except pystac.STACValidationError as e:
        logging.warning(
            "Validation failed for DatasetDto %s converted to a STAC Collection %s: %s", dto.uuid, collection, e
        )
        raise

    return collection



from typing import Any, List, Optional, Set

from ..resource import ResourceDto, ResourceSpec
from ..resource_registry import DEFAULT_RESOURCE_TYPE_REGISTRY
from .distribution import Distribution
from .observable import ObservableDto
from odp.util.cheapdantic import Field


class DataCollectionSpec(ResourceSpec):
    distribution: Optional[Distribution] = None
    """Information on how the dataset was distributed"""

    tags: Set[str] = Field(default_factory=set)
    """Tags for the dataset"""

    facets: Optional[dict[str, Any]] = None
    """Facets for the dataset"""

    observables: List[ObservableDto] = Field(default_factory=list)
    """Observables for the data collection"""


class DataCollectionDto(ResourceDto):
    _kind: str = "catalog.hubocean.io/dataCollection"
    _version: str = "v1alpha1"

    spec: DataCollectionSpec


DEFAULT_RESOURCE_TYPE_REGISTRY.add(DataCollectionDto)



from math import log2

IEC_UNITS = ["KiB", "MiB", "GiB", "TiB"]


def size2human(size: int) -> str:
    if size == 0:
        return "0B"
    p = int(log2(size) // 10.0)

    if p < 1:
        return f"{size}B"
    if p > len(IEC_UNITS):
        p = len(IEC_UNITS)
    converted_size = size / 1024**p
    return f"{converted_size:.1f}{IEC_UNITS[p - 1]}"



from typing import Union

import requests

from ..util.cheapdantic import BaseModel
from .http_client import OdpHttpClient
from odp.client.exc import OdpUnauthorizedError
from odp.dto.catalog.dataset import DatasetDto
from odp.dto.resource import ResourceDto


class OdpIamClient(BaseModel):
    http_client: OdpHttpClient

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def share_with_hubocean_internal(self, ref: Union[ResourceDto, DatasetDto]) -> str:
        """Share a resource with HubOcean Internal group as editor.

        Args:
            ref: Resource reference
            subject_id: Subject ID
            role: Role to assign to the subject
        """
        if not (isinstance(ref, ResourceDto) or isinstance(ref, DatasetDto)):
            raise ValueError("ref is not of type ResourceDto")

        role = "2"  # Editor
        subject_id = "0ee75a5a-6fcc-47db-8d0c-8f61f0641126"  # HubOcean Internal group ID
        params = {
            "object": {"id": str(ref.metadata.uuid), "kind": str(ref.get_kind())},
            "role": role,
            "subject": {"id": subject_id, "type": "group"},
        }
        path = "/api/permissions/v1/resources/relationships/"
        res = self.http_client.post(f"{path}", content=params)

        try:
            res.raise_for_status()
        except requests.HTTPError:
            if res.status_code == 409:
                return "The resource is already shared with HubOcean Internal group"
            elif res.status_code == 401:
                raise OdpUnauthorizedError("Unauthorized access")
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")
        return "Success"



from typing import List, Union
from uuid import UUID

import requests

from odp.client.exc import OdpValidationError
from odp.client.http_client import OdpHttpClient
from odp.dto.catalogv2.catalog import (
    CreateDataCollectionRequest,
    CreateDatasetRequest,
    DataCollection,
    Dataset,
    UpdateDataCollectionRequest,
    UpdateDatasetRequest,
)
from odp.dto.catalogv2.search import (
    AutoCompletionGroupedResult,
    AutoCompletionRequest,
    AutoCompletionResult,
    AutoCompletionResultItem,
    AutoCompletionResultType,
    AutoCompletionScore,
    CollectionSearchResultPayload,
    DatasetSearchResultPayload,
    PayloadType,
    ResultItem,
    SearchRequest,
    SearchResult,
    SearchScore,
)
from odp.util.cheapdantic.cheapdantic import BaseModel


class CatalogClient(BaseModel):

    """Client for interacting with Datasets."""

    http_client: OdpHttpClient
    catalog_base_url: str = "/api/catalog/v2"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def create_dataset(
        self,
        request: CreateDatasetRequest,
    ) -> Dataset:
        """Create a dataset.

        Args:
            request: The dataset creation request object

        Returns:
            The created dataset

        Raises:
            OdpValidationError: If the request is invalid
            requests.HTTPError: For other unexpected HTTP errors
        """
        res = self.http_client.post(self.catalog_base_url + "/datasets", content=request)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            if res.status_code == 500:
                raise requests.HTTPError(f"Internal server error ({res.status_code}): {res.text}") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")

        return Dataset(**res.json())

    def update_dataset(
        self,
        dataset_id: str,
        request: UpdateDatasetRequest,
    ) -> Dataset:
        """Update a dataset.

        Args:
            dataset_id: The ID of the dataset to update.
            request: The dataset update request object.

        Returns:
            The updated dataset.

        Raises:
            OdpValidationError: If the request is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/datasets/{dataset_id}"
        res = self.http_client.put(url, content=request)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            if res.status_code == 500:
                raise requests.HTTPError(f"Internal server error ({res.status_code}): {res.text}") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")

        return Dataset(**res.json())

    def get_datasets(self) -> List[Dataset]:
        """Fetch all datasets accessible to the user.

        Returns:
            A list of Dataset objects.

        Raises:
            requests.HTTPError: For unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/datasets"
        res = self.http_client.get(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 500:
                raise requests.HTTPError(f"Internal server error ({res.status_code}): {res.text}") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")

        return [Dataset(**item) for item in res.json()]

    def get_dataset_by_id(self, dataset_id: Union[str, UUID]) -> Dataset:
        """Fetch a dataset by ID.

        Args:
            dataset_id: The ID of the dataset to fetch.

        Returns:
            A Dataset object.

        Raises:
            OdpValidationError: If the ID is malformed or request is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/datasets/{dataset_id}"
        res = self.http_client.get(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return Dataset(**res.json())

    def delete_dataset(self, dataset_id: Union[str, UUID]) -> None:
        """Delete a dataset by ID.

        Args:
            dataset_id: The ID of the dataset to delete.

        Raises:
            OdpValidationError: If the ID is malformed or request is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/datasets/{dataset_id}"
        res = self.http_client.delete(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

    def archive_dataset(self, dataset_id: Union[str, UUID]) -> Dataset:
        """Archive a dataset by ID.

        Args:
            dataset_id: The ID of the dataset to archive.

        Returns:
            The updated Dataset object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/datasets/{dataset_id}/archive"
        res = self.http_client.patch(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return Dataset(**res.json())

    def make_dataset_public(self, dataset_id: Union[str, UUID]) -> Dataset:
        """Make a dataset public by ID.

        Args:
            dataset_id: The ID of the dataset to make public.

        Returns:
            The updated Dataset object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/datasets/{dataset_id}/visibility/public"
        res = self.http_client.patch(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return Dataset(**res.json())

    def make_dataset_private(self, dataset_id: Union[str, UUID]) -> Dataset:
        """Make a dataset private by ID.

        Args:
            dataset_id: The ID of the dataset to make private.

        Returns:
            The updated Dataset object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/datasets/{dataset_id}/visibility/private"
        res = self.http_client.patch(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return Dataset(**res.json())

    def create_data_collection(self, request: CreateDataCollectionRequest) -> DataCollection:
        """Create a data collection.

        Args:
            request: The data collection creation request object.

        Returns:
            The created DataCollection object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections"
        res = self.http_client.post(url, content=request)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return DataCollection(**res.json())

    def get_data_collections(self) -> List[DataCollection]:
        """Fetch all data collections accessible to the user.

        Returns:
            A list of DataCollection objects.

        Raises:
            requests.HTTPError: For unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections"
        res = self.http_client.get(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return [DataCollection(**item) for item in res.json()]

    def get_data_collection_by_id(self, data_collection_id: Union[str, UUID]) -> DataCollection:
        """Fetch a data collection by ID.

        Args:
            data_collection_id: The ID of the data collection to fetch.

        Returns:
            A DataCollection object.

        Raises:
            OdpValidationError: If the input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections/{data_collection_id}"
        res = self.http_client.get(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return DataCollection(**res.json())

    def update_data_collection(
        self,
        data_collection_id: Union[str, UUID],
        request: UpdateDataCollectionRequest,
    ) -> DataCollection:
        """Update a data collection.

        Args:
            data_collection_id: The ID of the data collection to update.
            request: The update request object.

        Returns:
            The updated DataCollection object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections/{data_collection_id}"
        res = self.http_client.put(url, content=request)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return DataCollection(**res.json())

    def delete_data_collection(self, data_collection_id: Union[str, UUID]) -> None:
        """Delete a data collection by ID.

        Args:
            data_collection_id: The ID of the data collection to delete.

        Raises:
            OdpValidationError: If the input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections/{data_collection_id}"
        res = self.http_client.delete(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

    def make_data_collection_public(self, data_collection_id: Union[str, UUID]) -> DataCollection:
        """Make a data collection public by ID.

        Args:
            data_collection_id: The ID of the data collection to make public.

        Returns:
            The updated DataCollection object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections/{data_collection_id}/visibility/public"
        res = self.http_client.patch(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return DataCollection(**res.json())

    def make_data_collection_private(self, data_collection_id: Union[str, UUID]) -> DataCollection:
        """Make a data collection private by ID.

        Args:
            data_collection_id: The ID of the data collection to make private.

        Returns:
            The updated DataCollection object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections/{data_collection_id}/visibility/private"
        res = self.http_client.patch(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return DataCollection(**res.json())

    def add_dataset_to_data_collection(
        self,
        data_collection_id: Union[str, UUID],
        dataset_id: Union[str, UUID],
    ) -> DataCollection:
        """Add a dataset to a data collection.

        Args:
            data_collection_id: The ID of the data collection.
            dataset_id: The ID of the dataset to add.

        Returns:
            The updated DataCollection object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections/{data_collection_id}/datasets/{dataset_id}"
        res = self.http_client.post(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return DataCollection(**res.json())

    def remove_dataset_from_data_collection(
        self,
        data_collection_id: Union[str, UUID],
        dataset_id: Union[str, UUID],
    ) -> DataCollection:
        """Remove a dataset from a data collection.

        Args:
            data_collection_id: The ID of the data collection.
            dataset_id: The ID of the dataset to remove.

        Returns:
            The updated DataCollection object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/data-collections/{data_collection_id}/datasets/{dataset_id}"
        res = self.http_client.delete(url)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        return DataCollection(**res.json())

    def search(self, request: SearchRequest) -> SearchResult:
        """Perform a search with typed payloads."""
        url = f"{self.catalog_base_url}/search"
        res = self.http_client.post(url, content=request)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        raw = res.json()
        search_result = SearchResult(
            total=raw["total"],
            top_score=SearchScore(**raw["top_score"]) if raw.get("top_score") else None,
            match=raw["match"],
            results=[],
        )

        for item in raw["results"]:
            payload_type = item.get("payload_type")

            if payload_type == PayloadType.DATASET:
                payload = DatasetSearchResultPayload(**item["payload"])
            elif payload_type == PayloadType.DATA_COLLECTION:
                payload = CollectionSearchResultPayload(**item["payload"])
            else:
                payload = item["payload"]

            result_item = ResultItem(
                id=item["id"],
                title=item["title"],
                description=item["description"],
                highlights=item.get("highlights", []),
                match_type=item["match_type"],
                payload_type=payload_type,
                score=SearchScore(**item["score"]),
                payload=payload,
            )
            search_result.results.append(result_item)

        return search_result

    def autocomplete(self, request: AutoCompletionRequest) -> AutoCompletionResult:
        """Perform an autocomplete search.

        Args:
            request: The autocomplete request payload.

        Returns:
            An AutoCompletionResult object.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/autocomplete"
        res = self.http_client.post(url, content=request)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        raw = res.json()
        auto_completion_result = AutoCompletionResult(
            total=raw["total"],
            top_score=AutoCompletionScore(**raw["top_score"]) if raw.get("top_score") else None,
            results=[],
        )

        for item in raw["results"]:
            result_item = AutoCompletionResultItem(
                id=item["id"],
                type=AutoCompletionResultType(item["type"]),
                title=item["title"],
                description=item["description"],
                score=AutoCompletionScore(**item["score"]),
            )
            auto_completion_result.results.append(result_item)

        return auto_completion_result

    def autocomplete_grouped(self, request: AutoCompletionRequest) -> List[AutoCompletionGroupedResult]:
        """Perform a grouped autocomplete search.

        Args:
            request: The autocomplete request payload.

        Returns:
            A list of AutoCompletionGroupedResult objects.

        Raises:
            OdpValidationError: If the request input is invalid.
            requests.HTTPError: For other unexpected HTTP errors.
        """
        url = f"{self.catalog_base_url}/autocomplete/grouped"
        res = self.http_client.post(url, content=request)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}") from e

        raw = res.json()
        grouped_results = []

        for group_item in raw:
            result_items = [
                AutoCompletionResultItem(
                    id=item["id"],
                    type=AutoCompletionResultType(item["type"]),
                    title=item["title"],
                    description=item["description"],
                    score=AutoCompletionScore(**item["score"]),
                )
                for item in (group_item["result"].get("results") or [])
            ]

            result = AutoCompletionResult(
                total=group_item["result"]["total"],
                top_score=AutoCompletionScore(**group_item["result"]["top_score"])
                if group_item["result"].get("top_score")
                else None,
                results=result_items,
            )

            grouped_result = AutoCompletionGroupedResult(
                group=AutoCompletionResultType(group_item["group"]), result=result
            )

            grouped_results.append(grouped_result)

        return grouped_results



from datetime import datetime
from typing import Optional

from ..common import ContactInfo, License
from odp.util.cheapdantic import BaseModel


class Distribution(BaseModel):
    """Distribution information"""

    published_by: ContactInfo
    """Publisher information"""

    published_date: datetime
    """Date of first published"""

    website: str
    """Distribution website"""

    license: Optional[License] = None
    """Dataset license information"""



import datetime
import json
import logging
import types
from decimal import Decimal
from enum import Enum
from typing import Any, Callable, Dict, Type, Union, _GenericAlias
from uuid import UUID


class Field:
    def __init__(self, default_factory=None):
        self.default_factory = default_factory


class PrivateAttr:
    def __init__(self, val):
        self.val = val


class SecretStr:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "********"

    def __str__(self):
        return "********"

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __hash__(self):
        return hash(self.value)

    def get_secret_value(self):
        return self.value


class BaseModel:
    def __init__(self, **kwargs):
        mn = type(self).__name__
        attrs = _attrs(self.__class__)
        for name, typ in attrs.items():
            if name in kwargs:
                val = kwargs.get(name)
                logging.debug("field %s.%s in kwargs: %s", mn, name, val)
            else:
                logging.debug("field %s.%s not in kwargs", mn, name)
                val = getattr(self, name, None)
                if isinstance(val, Field):
                    val = val.default_factory()
                if isinstance(val, PrivateAttr):
                    setattr(self, name, val.val)
                    continue
            if typ == PrivateAttr:
                continue
            if typ == SecretStr:
                if not isinstance(val, SecretStr):
                    val = SecretStr(str(val))

            subtype = typ
            if isinstance(typ, types.GenericAlias):
                subtype = typ.__args__[0]
            if isinstance(typ, _GenericAlias):
                subtype = typ.__args__[0]

            if subtype == UUID:
                if val is None:
                    pass
                elif not isinstance(val, UUID):
                    val = UUID(str(val))

            if subtype == datetime.datetime:
                if val is None:
                    val = None
                elif not isinstance(val, datetime.datetime):
                    val = datetime.datetime.fromisoformat(str(val))

            if BaseModel in getattr(subtype, "__mro__", []):
                logging.debug("field %s.%s is BaseModel: %s", mn, name, typ)
                if isinstance(val, dict):
                    val = subtype(**val)
            elif not _is_type(val, typ) and not _is_type(val, Any):
                raise ValueError(f"field {mn}.{name} value {val} is not of type {typ} but of type {type(val)}")

            setattr(self, name, val)

    def json(self, exclude_unset=False):
        return json.dumps(self, cls=JSONCodec, skipkeys=exclude_unset)

    def dict(self, exclude_unset=False):
        res_dict = self.__dict__.copy()
        for k, v in self.__dict__.items():
            if isinstance(v, BaseModel):
                res_dict[k] = v.dict(exclude_unset)
            if k.startswith("_"):
                res_dict.pop(k)
            if exclude_unset and v is None:
                res_dict.pop(k)
        return res_dict

    def __str__(self):
        return str(self.json())

    def __repr__(self):
        return str(self.json())

    def __iter__(self):
        return iter(self.dict().items())

    def __eq__(self, other):
        if not isinstance(other, BaseModel):
            return False
        return self.json() == other.json()


def _attrs(cls):
    attrs = {}
    # check for order
    for x in cls.__mro__:
        if hasattr(x, "__annotations__"):
            for name, typ in x.__annotations__.items():
                attrs.setdefault(name, typ)
    return attrs


def _is_type(val, typ) -> bool:
    if isinstance(typ, list):
        # logging.info("_is_type(%s, list %s)", val, typ)
        for t in typ:
            if _is_type(val, t):
                return True
        return False
    if isinstance(typ, (types.GenericAlias, _GenericAlias)):  # supported in all python versions?
        if typ.__origin__ is Union:
            return any(_is_type(val, t) for t in typ.__args__)
        typ = typ.__origin__  # we only bother to check the main type, not the parametrized ones
    try:
        if typ is Any:
            return True
        if isinstance(val, typ):
            return True
    except Exception as e:
        if isinstance(val, SecretStr):
            logging.warning("is_type(%s, %s) failed: %s", "SensitiveInformation", typ, e)
        else:
            logging.warning("is_type(%s, %s) failed: %s", val, typ, e)
        return True
    return False


class JSONCodec(json.JSONEncoder, json.JSONDecoder):
    encoders_by_type: Dict[Type[Any], Callable[[Any], Any]] = {
        bytes: lambda o: o.decode(),
        datetime.date: lambda d: d.isoformat(),
        datetime.datetime: lambda dt: dt.isoformat(),
        datetime.time: lambda t: t.isoformat(),
        datetime.timedelta: lambda td: td.total_seconds(),
        Decimal: lambda x: int(x) if x.as_tuple().exponent >= 0 else float(x),
        Enum: lambda o: o.value,
        SecretStr: str,
        set: list,
        UUID: str,
    }

    def default(self, obj):
        if isinstance(obj, BaseModel):
            return obj.dict(exclude_unset=self.skipkeys)

        for typ, encoder in self.encoders_by_type.items():
            if isinstance(obj, typ):
                return encoder(obj)
        raise TypeError(f"Object of type '{obj.__class__.__name__}' is not JSON serializable")
